# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helper functions for converting the Scene object to some other object."""

from __future__ import annotations

import typing

import xarray as xr

from satpy.composites.core import enhance2dataset
from satpy.dataset import DataID

if typing.TYPE_CHECKING:
    from satpy.scene import Scene

def _get_dataarrays_from_identifiers(scn, identifiers):
    """Return a list of DataArray based on a single or list of identifiers.

    An identifier can be a DataID or a string with name of a valid DataID.
    """
    if isinstance(identifiers, (str, DataID)):
        identifiers = [identifiers]

    if identifiers is not None:
        dataarrays = [scn[ds] for ds in identifiers]
    else:
        dataarrays = [scn._datasets.get(ds) for ds in scn._wishlist]
        dataarrays = [dataarray for dataarray in dataarrays if dataarray is not None]
    return dataarrays


def to_geoviews(
        scn: Scene,
        gvtype: typing.Any | None = None,
        datasets: list | None = None,
        vdims: list[str] | None = None,
        dynamic: bool = False,
):
        """Convert satpy Scene to geoviews.

        Args:
            scn: Satpy Scene.
            gvtype:
                One of gv.Image, gv.LineContours, gv.FilledContours, gv.Points
                Default to :class:`geoviews.Image <geoviews.element.geo.Image>`.
                See Geoviews documentation for details.
            datasets: Limit included products to these datasets
            vdims:
                Value dimensions. See geoviews documentation for more information.
                If not given defaults to first data variable
            dynamic: Load and compute data on-the-fly during
                visualization. Default is ``False``. See
                https://holoviews.org/user_guide/Gridded_Datasets.html#working-with-xarray-data-types
                for more information. Has no effect when data to be visualized
                only has 2 dimensions (y/x or longitude/latitude) and doesn't
                require grouping via the Holoviews ``groupby`` function.

        Returns: geoviews object

        Todo:
            * better handling of projection information in datasets which are
              to be passed to geoviews

        """
        import geoviews as gv
        from cartopy import crs  # noqa
        if gvtype is None:
            gvtype = gv.Image

        ds = scn.to_xarray_dataset(datasets)

        if vdims is None:
            # by default select first data variable as display variable
            vdims = ds.data_vars[list(ds.data_vars.keys())[0]].name

        if hasattr(ds, "area") and hasattr(ds.area, "to_cartopy_crs"):
            dscrs = ds.area.to_cartopy_crs()
            gvds = gv.Dataset(ds, crs=dscrs)
        else:
            gvds = gv.Dataset(ds)

        # holoviews produces a log warning if you pass groupby arguments when groupby isn't used
        groupby_kwargs = {"dynamic": dynamic} if gvds.ndims != 2 else {}
        if "latitude" in ds.coords:
            gview = gvds.to(gv.QuadMesh, kdims=["longitude", "latitude"],
                            vdims=vdims, **groupby_kwargs)
        else:
            gview = gvds.to(gvtype, kdims=["x", "y"], vdims=vdims,
                            **groupby_kwargs)

        return gview

def to_hvplot(scn, datasets=None, *args, **kwargs):
        """Convert satpy Scene to Hvplot. The method could not be used with composites of swath data.

        Args:
            scn (satpy.scene.Scene): Satpy Scene.
            datasets (list): Limit included products to these datasets.
            args: Arguments coming from hvplot
            kwargs: hvplot options dictionary.

        Returns:
            hvplot object that contains within it the plots of datasets list.
            As default it contains all Scene datasets plots and a plot title
            is shown.

        Example usage::

           scene_list = ['ash','IR_108']
           scn = Scene()
           scn.load(scene_list)
           scn = scn.resample('eurol')
           plot = scn.to_hvplot(datasets=scene_list)
           plot.ash+plot.IR_108
        """

        def _get_crs(xarray_ds):
            return xarray_ds.area.to_cartopy_crs()

        def _get_timestamp(xarray_ds):
            time = xarray_ds.attrs["start_time"]
            return time.strftime("%Y %m %d -- %H:%M UTC")

        def _get_units(xarray_ds, variable):
            return xarray_ds[variable].attrs["units"]

        def _plot_rgb(xarray_ds, variable, **defaults):
            img = enhance2dataset(xarray_ds[variable])
            return img.hvplot.rgb(bands="bands", title=title,
                                  clabel="", **defaults)

        def _plot_quadmesh(xarray_ds, variable, **defaults):
            return xarray_ds[variable].hvplot.quadmesh(
                clabel=f"[{_get_units(xarray_ds,variable)}]", title=title,
                **defaults)

        import hvplot.xarray as hvplot_xarray  # noqa
        from holoviews import Overlay

        plot = Overlay()
        xarray_ds = scn.to_xarray_dataset(datasets)

        if hasattr(xarray_ds, "area") and hasattr(xarray_ds.area, "to_cartopy_crs"):
            ccrs = _get_crs(xarray_ds)
            defaults={"x":"x","y":"y"}
        else:
            ccrs = None
            defaults={"x":"longitude","y":"latitude"}

        if datasets is None:
            datasets = list(xarray_ds.keys())

        defaults.update(data_aspect=1, project=True, geo=True,
                        crs=ccrs, projection=ccrs, rasterize=True,
                        coastline="110m", cmap="Plasma", responsive=True,
                        dynamic=False, framewise=True,colorbar=False,
                        global_extent=False, xlabel="Longitude",
                        ylabel="Latitude")

        defaults.update(kwargs)

        for element in datasets:
            title = f"{element} @ {_get_timestamp(xarray_ds)}"
            if xarray_ds[element].shape[0] == 3:
                plot[element] = _plot_rgb(xarray_ds, element, **defaults)
            else:
                plot[element] = _plot_quadmesh(xarray_ds, element, **defaults)

        return plot

def to_xarray(scn,
              datasets=None,  # DataID
              header_attrs=None,
              exclude_attrs=None,
              flatten_attrs=False,
              pretty=True,
              include_lonlats=True,
              epoch=None,
              include_orig_name=True,
              numeric_name_prefix="CHANNEL_"):
    """Merge all xr.DataArray(s) of a satpy.Scene to a CF-compliant xarray object.

    If all Scene DataArrays are on the same area, it returns an xr.Dataset.
    If Scene DataArrays are on different areas, currently it fails, although
    in future we might return a DataTree object, grouped by area.

    Args:
        scn (satpy.scene.Scene): Satpy Scene.
        datasets (Iterable, Optional): List of Satpy Scene datasets to include in
            the output xr.Dataset. Elements can be string name, a wavelength as a
            number, a DataID, or DataQuery object. If None (the default), it
            includes all loaded Scene datasets.
        header_attrs: Global attributes of the output xr.Dataset.
        epoch (str, Optional): Reference time for encoding the time coordinates
            (if available). Format example: "seconds since 1970-01-01 00:00:00".
            If None, the default reference time is retrieved using
            "from satpy.cf_writer import EPOCH".
        flatten_attrs (bool, Optional): If True, flatten dict-type attributes.
        exclude_attrs (list, Optional): List of xr.DataArray attribute names to
            be excluded.
        include_lonlats (bool, Optional): If True, includes 'latitude' and
            'longitude' coordinates. If the 'area' attribute is a SwathDefinition,
            it always includes latitude and longitude coordinates.
        pretty (bool, Optional): Don't modify coordinate names, if possible. Makes
            the file prettier, but possibly less consistent.
        include_orig_name (bool, Optional): Include the original dataset name as a
            variable attribute in the xr.Dataset.
        numeric_name_prefix (str, Optional): Prefix to add to each variable with
            name starting with a digit. Use '' or None to leave this out.

    Returns:
        xarray.Dataset: A CF-compliant xr.Dataset

    """
    from satpy.cf.datasets import collect_cf_datasets

    # Get list of DataArrays
    if datasets is None:
        datasets = list(scn.keys())  # list all loaded DataIDs
    list_dataarrays = _get_dataarrays_from_identifiers(scn, datasets)

    # Check that some DataArray could be returned
    if len(list_dataarrays) == 0:
        return xr.Dataset()

    # Collect xr.Dataset for each group
    grouped_datasets, header_attrs = collect_cf_datasets(list_dataarrays=list_dataarrays,
                                                         header_attrs=header_attrs,
                                                         exclude_attrs=exclude_attrs,
                                                         flatten_attrs=flatten_attrs,
                                                         pretty=pretty,
                                                         include_lonlats=include_lonlats,
                                                         epoch=epoch,
                                                         include_orig_name=include_orig_name,
                                                         numeric_name_prefix=numeric_name_prefix,
                                                         groups=None)
    if len(grouped_datasets) == 1:
        ds = grouped_datasets[None]
        return ds
    else:
        msg = """The Scene object contains datasets with different areas.
                  Resample the Scene to have matching dimensions using i.e. scn.resample(resampler="native") """
        raise NotImplementedError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.core.netcdf module."""

import os

import numpy as np
import pytest

from satpy.readers.core.netcdf import NetCDF4FileHandler, choose_accessor_from_engine


class FakeNetCDF4FileHandler(NetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler for reader tests to use."""

    def __init__(self, filename, filename_info, filetype_info,
                 auto_maskandscale=False, xarray_kwargs=None,
                 cache_var_size=0, cache_handle=False, extra_file_content=None):
        """Get fake file content from 'get_test_content'."""
        # unused kwargs from the real file handler
        del auto_maskandscale
        del xarray_kwargs
        del cache_var_size
        del cache_handle
        super(NetCDF4FileHandler, self).__init__(filename, filename_info, filetype_info)
        self.file_content = self.get_test_content(filename, filename_info, filetype_info)
        if extra_file_content:
            self.file_content.update(extra_file_content)
        self.engine = "netcdf4"
        self.accessor = choose_accessor_from_engine(self.engine)

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content.

        Args:
            filename (str): input filename
            filename_info (dict): Dict of metadata pulled from filename
            filetype_info (dict): Dict of metadata from the reader's yaml config for this file type

        Returns: dict of file content with keys like:

            - 'dataset'
            - '/attr/global_attr'
            - 'dataset/attr/global_attr'
            - 'dataset/shape'
            - 'dataset/dimensions'
            - '/dimension/my_dim'

        """
        raise NotImplementedError("Fake File Handler subclass must implement 'get_test_content'")


@pytest.fixture
def netcdf_file(tmp_path):
    """Create a test NetCDF4 file."""
    from netCDF4 import Dataset
    filename = tmp_path / "test.nc"
    with Dataset(filename, "w") as nc:
        # Create dimensions
        nc.createDimension("rows", 10)
        nc.createDimension("cols", 100)

        # Create Group
        g1 = nc.createGroup("test_group")

        # Add datasets
        ds1_f = g1.createVariable("ds1_f", np.float32,
                                  dimensions=("rows", "cols"))
        ds1_f[:] = np.arange(10. * 100).reshape((10, 100))
        ds1_i = g1.createVariable("ds1_i", np.int32,
                                  dimensions=("rows", "cols"))
        ds1_i[:] = np.arange(10 * 100).reshape((10, 100))
        ds2_f = nc.createVariable("ds2_f", np.float32,
                                  dimensions=("rows", "cols"))
        ds2_f[:] = np.arange(10. * 100).reshape((10, 100))
        ds2_i = nc.createVariable("ds2_i", np.int32,
                                  dimensions=("rows", "cols"))
        ds2_i[:] = np.arange(10 * 100).reshape((10, 100))
        ds2_s = nc.createVariable("ds2_s", np.int8,
                                  dimensions=("rows",))
        ds2_s[:] = np.arange(10)
        ds2_sc = nc.createVariable("ds2_sc", np.int8, dimensions=())
        ds2_sc[:] = 42

        # Add attributes
        nc.test_attr_str = "test_string"
        nc.test_attr_int = 0
        nc.test_attr_float = 1.2
        nc.test_attr_str_arr = np.array(b"test_string2")
        g1.test_attr_str = "test_string"
        g1.test_attr_int = 0
        g1.test_attr_float = 1.2
        for d in [ds1_f, ds1_i, ds2_f, ds2_i]:
            d.test_attr_str = "test_string"
            d.test_attr_int = 0
            d.test_attr_float = 1.2
    return filename

class TestNetCDF4FileHandler:
    """Test NetCDF4 File Handler Utility class."""

    def test_all_basic(self, netcdf_file):
        """Test everything about the NetCDF4 class."""
        import xarray as xr

        from satpy.readers.core.netcdf import NetCDF4FileHandler
        file_handler = NetCDF4FileHandler(netcdf_file, {}, {})

        assert file_handler["/dimension/rows"] == 10
        assert file_handler["/dimension/cols"] == 100

        for ds in ("test_group/ds1_f", "test_group/ds1_i", "ds2_f", "ds2_i"):
            assert file_handler[ds].dtype == (np.float32 if ds.endswith("f") else np.int32)
            assert file_handler[ds + "/shape"] == (10, 100)
            assert file_handler[ds + "/dimensions"] == ("rows", "cols")
            assert file_handler[ds + "/attr/test_attr_str"] == "test_string"
            assert file_handler[ds + "/attr/test_attr_int"] == 0
            assert file_handler[ds + "/attr/test_attr_float"] == 1.2

        test_group = file_handler["test_group"]
        assert test_group["ds1_i"].shape == (10, 100)
        assert test_group["ds1_i"].dims == ("rows", "cols")

        assert file_handler["/attr/test_attr_str"] == "test_string"
        assert file_handler["/attr/test_attr_str_arr"] == "test_string2"
        assert file_handler["/attr/test_attr_int"] == 0
        assert file_handler["/attr/test_attr_float"] == 1.2

        global_attrs = {
            "test_attr_str": "test_string",
            "test_attr_str_arr": "test_string2",
            "test_attr_int": 0,
            "test_attr_float": 1.2
            }
        assert file_handler["/attrs"] == global_attrs

        assert isinstance(file_handler.get("ds2_f")[:], xr.DataArray)
        assert file_handler.get("fake_ds") is None
        assert file_handler.get("fake_ds", "test") == "test"

        assert ("ds2_f" in file_handler) is True
        assert ("fake_ds" in file_handler) is False
        assert file_handler.file_handle is None
        assert file_handler["ds2_sc"] == 42

    def test_listed_variables(self, netcdf_file):
        """Test that only listed variables/attributes area collected."""
        from satpy.readers.core.netcdf import NetCDF4FileHandler

        filetype_info = {
            "required_netcdf_variables": [
                "test_group/attr/test_attr_str",
                "attr/test_attr_str",
            ]
        }
        file_handler = NetCDF4FileHandler(netcdf_file, {}, filetype_info)
        assert len(file_handler.file_content) == 2
        assert "test_group/attr/test_attr_str" in file_handler.file_content
        assert "attr/test_attr_str" in file_handler.file_content

    def test_listed_variables_with_composing(self, netcdf_file):
        """Test that composing for listed variables is performed."""
        from satpy.readers.core.netcdf import NetCDF4FileHandler

        filetype_info = {
            "required_netcdf_variables": [
                "test_group/{some_parameter}/attr/test_attr_str",
                "test_group/attr/test_attr_str",
            ],
            "variable_name_replacements": {
                "some_parameter": [
                    "ds1_f",
                    "ds1_i",
                ],
                "another_parameter": [
                    "not_used"
                ],
            }
        }
        file_handler = NetCDF4FileHandler(netcdf_file, {}, filetype_info)
        assert len(file_handler.file_content) == 3
        assert "test_group/ds1_f/attr/test_attr_str" in file_handler.file_content
        assert "test_group/ds1_i/attr/test_attr_str" in file_handler.file_content
        assert not any("not_used" in var for var in file_handler.file_content)
        assert not any("some_parameter" in var for var in file_handler.file_content)
        assert not any("another_parameter" in var for var in file_handler.file_content)
        assert "test_group/attr/test_attr_str" in file_handler.file_content

    def test_caching(self, netcdf_file):
        """Test that caching works as intended."""
        from satpy.readers.core.netcdf import NetCDF4FileHandler
        h = NetCDF4FileHandler(netcdf_file, {}, {}, cache_var_size=1000,
                               cache_handle=True)
        assert h.file_handle is not None
        assert h.file_handle.isopen()

        assert sorted(h.cached_file_content.keys()) == ["ds2_s", "ds2_sc"]
        # with caching, these tests access different lines than without
        np.testing.assert_array_equal(h["ds2_s"], np.arange(10))
        np.testing.assert_array_equal(h["test_group/ds1_i"],
                                      np.arange(10 * 100).reshape((10, 100)))
        # check that root variables can still be read from cached file object,
        # even if not cached themselves
        np.testing.assert_array_equal(
                h["ds2_f"],
                np.arange(10. * 100).reshape((10, 100)))
        h.__del__()
        assert not h.file_handle.isopen()

    def test_filenotfound(self):
        """Test that error is raised when file not found."""
        from satpy.readers.core.netcdf import NetCDF4FileHandler

        # NOTE: Some versions of NetCDF C report unknown file format on Windows
        with pytest.raises(IOError, match=".*(No such file or directory|Unknown file format).*"):
            NetCDF4FileHandler("/thisfiledoesnotexist.nc", {}, {})

    def test_get_and_cache_npxr_is_xr(self, netcdf_file):
        """Test that get_and_cache_npxr() returns xr.DataArray."""
        import xarray as xr

        from satpy.readers.core.netcdf import NetCDF4FileHandler
        file_handler = NetCDF4FileHandler(netcdf_file, {}, {}, cache_handle=True)

        data = file_handler.get_and_cache_npxr("test_group/ds1_f")
        assert isinstance(data, xr.DataArray)

    def test_get_and_cache_npxr_data_is_cached(self, netcdf_file):
        """Test that the data are cached when get_and_cache_npxr() is called."""
        from satpy.readers.core.netcdf import NetCDF4FileHandler

        file_handler = NetCDF4FileHandler(netcdf_file, {}, {}, cache_handle=True)
        data = file_handler.get_and_cache_npxr("test_group/ds1_f")

        # Delete the dataset from the file content dict, it should be available from the cache
        del file_handler.file_content["test_group/ds1_f"]
        data2 = file_handler.get_and_cache_npxr("test_group/ds1_f")
        assert np.all(data == data2)

class TestNetCDF4FsspecFileHandler:
    """Test the remote reading class."""

    def test_default_to_netcdf4_lib(self):
        """Test that the NetCDF4 backend is used by default."""
        import tempfile

        import h5py

        from satpy.readers.core.netcdf import NetCDF4FsspecFileHandler

        with tempfile.TemporaryDirectory() as tmpdir:
            # Create an empty HDF5
            fname = os.path.join(tmpdir, "test.nc")
            fid = h5py.File(fname, "w")
            fid.close()

            fh = NetCDF4FsspecFileHandler(fname, {}, {})
            assert fh.accessor.engine == "netcdf4"

    def test_use_h5netcdf_for_file_not_accessible_locally(self):
        """Test that h5netcdf is used for files that are not accesible locally."""
        from unittest.mock import patch

        fname = "s3://bucket/object.nc"

        with patch("h5netcdf.File") as h5_file:
            with patch("satpy.readers.core.netcdf.open_file_or_filename"):
                from satpy.readers.core.netcdf import NetCDF4FsspecFileHandler

                fh = NetCDF4FsspecFileHandler(fname, {}, {})
                h5_file.assert_called()
                assert fh.accessor.engine == "h5netcdf"

    @pytest.mark.parametrize("engine", ["netcdf4", "h5netcdf"])
    def test_netcdf_engines(self, netcdf_file, engine):
        """Test that h5netcdf engine is used."""
        from satpy.readers.core.netcdf import NetCDF4FsspecFileHandler

        fh = NetCDF4FsspecFileHandler(netcdf_file, {}, {}, engine=engine)
        assert fh.accessor.engine == engine
        np.testing.assert_array_equal(
                fh["ds2_f"],
                np.arange(10. * 100).reshape((10, 100)))


NC_ATTRS = {
    "standard_name": "test_data",
    "scale_factor": 0.01,
    "add_offset": 0}

def test_get_data_as_xarray_netcdf4(tmp_path):
    """Test getting xr.DataArray from netcdf4 variable."""
    import numpy as np

    from satpy.readers.core.netcdf import get_data_as_xarray

    data = np.array([1, 2, 3])
    fname = tmp_path / "test.nc"
    dset = _write_test_netcdf4(fname, data)

    res = get_data_as_xarray(dset["test_data"])
    np.testing.assert_equal(res.data, data)
    assert res.attrs == NC_ATTRS


def test_get_data_as_xarray_scalar_netcdf4(tmp_path):
    """Test getting scalar xr.DataArray from netcdf4 variable."""
    import numpy as np

    from satpy.readers.core.netcdf import get_data_as_xarray

    data = 1
    fname = tmp_path / "test.nc"
    dset = _write_test_netcdf4(fname, data)

    res = get_data_as_xarray(dset["test_data"])
    np.testing.assert_equal(res.data, np.array(data))
    assert res.attrs == NC_ATTRS


def _write_test_netcdf4(fname, data):
    import netCDF4 as nc

    dset = nc.Dataset(fname, "w")
    try:
        dset.createDimension("y", data.size)
        dims = ("y",)
    except AttributeError:
        dims = ()
    var = dset.createVariable("test_data", "uint8", dims)
    var[:] = data
    var.setncatts(NC_ATTRS)
    # Turn off automatic scale factor and offset handling
    dset.set_auto_maskandscale(False)

    return dset


def test_get_data_as_xarray_h5netcdf(tmp_path):
    """Test getting xr.DataArray from h5netcdf variable."""
    import numpy as np

    from satpy.readers.core.netcdf import get_data_as_xarray

    data = np.array([1, 2, 3])
    fname = tmp_path / "test.nc"
    fid = _write_test_h5netcdf(fname, data)

    res = get_data_as_xarray(fid["test_data"])
    np.testing.assert_equal(res.data, data)
    assert res.attrs == NC_ATTRS


def _write_test_h5netcdf(fname, data):
    import h5netcdf

    fid = h5netcdf.File(fname, "w")
    try:
        fid.dimensions = {"y": data.size}
        dims = ("y",)
    except AttributeError:
        dims = ()
    var = fid.create_variable("test_data", dims, "uint8", data=data)
    for key in NC_ATTRS:
        var.attrs[key] = NC_ATTRS[key]

    return fid


def test_get_data_as_xarray_scalar_h5netcdf(tmp_path):
    """Test getting xr.DataArray from h5netcdf variable."""
    import numpy as np

    from satpy.readers.core.netcdf import get_data_as_xarray

    data = 1
    fname = tmp_path / "test.nc"
    fid = _write_test_h5netcdf(fname, data)

    res = get_data_as_xarray(fid["test_data"])
    np.testing.assert_equal(res.data, np.array(data))
    assert res.attrs == NC_ATTRS



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Utilities for various satpy tests."""

import datetime as dt
import os
from collections.abc import Mapping
from contextlib import contextmanager
from typing import Any
from unittest import mock

import dask.array as da
import numpy as np
from pyresample import create_area_def
from pyresample.geometry import BaseDefinition, SwathDefinition
from xarray import DataArray

from satpy import Scene
from satpy.composites.core import GenericCompositor, IncompatibleAreas
from satpy.dataset import DataID, DataQuery
from satpy.dataset.dataid import default_id_keys_config, minimal_default_keys_config
from satpy.modifiers import ModifierBase
from satpy.readers.core.file_handlers import BaseFileHandler

FAKE_FILEHANDLER_START = dt.datetime(2020, 1, 1, 0, 0, 0)
FAKE_FILEHANDLER_END = dt.datetime(2020, 1, 1, 1, 0, 0)

RANDOM_GEN = np.random.default_rng()


def make_dataid(**items):
    """Make a DataID with default keys."""
    return DataID(default_id_keys_config, **items)


def make_cid(**items):
    """Make a DataID with a minimal set of keys to id composites."""
    return DataID(minimal_default_keys_config, **items)


def make_dsq(**items):
    """Make a dataset query."""
    return DataQuery(**items)


def spy_decorator(method_to_decorate):
    """Fancy decorator to wrap an object while still calling it.

    See https://stackoverflow.com/a/41599695/433202

    """
    tmp_mock = mock.MagicMock()

    def wrapper(self, *args, **kwargs):
        tmp_mock(*args, **kwargs)
        return method_to_decorate(self, *args, **kwargs)

    wrapper.mock = tmp_mock
    return wrapper


def convert_file_content_to_data_array(file_content, attrs=tuple(),
                                       dims=("z", "y", "x")):
    """Help old reader tests that still use numpy arrays.

    A lot of old reader tests still use numpy arrays and depend on the
    "var_name/attr/attr_name" convention established before Satpy used xarray
    and dask. While these conventions are still used and should be supported,
    readers need to use xarray DataArrays instead.

    If possible, new tests should be based on pure DataArray objects instead
    of the "var_name/attr/attr_name" style syntax provided by the utility
    file handlers.

    Args:
        file_content (dict): Dictionary of string file keys to fake file data.
        attrs (Iterable): Series of attributes to copy to DataArray object from
            file content dictionary. Defaults to no attributes.
        dims (Iterable): Dimension names to use for resulting DataArrays.
            The second to last dimension is used for 1D arrays, so for
            dims of ``('z', 'y', 'x')`` this would use ``'y'``. Otherwise, the
            dimensions are used starting with the last, so 2D arrays are
            ``('y', 'x')``
            Dimensions are used in reverse order so the last dimension
            specified is used as the only dimension for 1D arrays and the
            last dimension for other arrays.

    """
    for key, val in file_content.items():
        da_attrs = {}
        for a in attrs:
            if key + "/attr/" + a in file_content:
                da_attrs[a] = file_content[key + "/attr/" + a]

        if isinstance(val, np.ndarray):
            val = da.from_array(val, chunks=4096)
            if val.ndim == 1:
                da_dims = dims[-2]
            elif val.ndim > 1:
                da_dims = tuple(dims[-val.ndim:])
            else:
                da_dims = None

            file_content[key] = DataArray(val, dims=da_dims, attrs=da_attrs)


def _filter_datasets(all_ds, names_or_ids):
    """Help filtering DataIDs by name or DataQuery."""
    # DataID will match a str to the name
    # need to separate them out
    str_filter = [ds_name for ds_name in names_or_ids if isinstance(ds_name, str)]
    id_filter = [ds_id for ds_id in names_or_ids if not isinstance(ds_id, str)]
    for ds_id in all_ds:
        if ds_id in id_filter or ds_id["name"] in str_filter:
            yield ds_id


def _swath_def_of_data_arrays(rows, cols):
    return SwathDefinition(
        DataArray(da.zeros((rows, cols)), dims=("y", "x")),
        DataArray(da.zeros((rows, cols)), dims=("y", "x")),
    )


class FakeModifier(ModifierBase):
    """Act as a modifier that performs different modifications."""

    def _handle_res_change(self, datasets, info):
        # assume this is used on the 500m version of ds5
        info["resolution"] = 250
        rep_data_arr = datasets[0]
        y_size = rep_data_arr.sizes["y"]
        x_size = rep_data_arr.sizes["x"]
        data = da.zeros((y_size * 2, x_size * 2))
        if isinstance(rep_data_arr.attrs["area"], SwathDefinition):
            area = _swath_def_of_data_arrays(y_size * 2, x_size * 2)
            info["area"] = area
        else:
            raise NotImplementedError("'res_change' modifier can't handle "
                                      "AreaDefinition changes yet.")
        return data

    def __call__(self, datasets, optional_datasets=None, **kwargs):
        """Modify provided data depending on the modifier name and input data."""
        if self.attrs["optional_prerequisites"]:
            for opt_dep in self.attrs["optional_prerequisites"]:
                opt_dep_name = opt_dep if isinstance(opt_dep, str) else opt_dep.get("name", "")
                if "NOPE" in opt_dep_name or "fail" in opt_dep_name:
                    continue
                assert optional_datasets is not None
                assert len(optional_datasets)
        resolution = datasets[0].attrs.get("resolution")
        mod_name = self.attrs["modifiers"][-1]
        data = datasets[0].data
        i = datasets[0].attrs.copy()
        if mod_name == "res_change" and resolution is not None:
            data = self._handle_res_change(datasets, i)
        elif "incomp_areas" in mod_name:
            raise IncompatibleAreas(
                "Test modifier 'incomp_areas' always raises IncompatibleAreas")
        self.apply_modifier_info(datasets[0].attrs, i)
        return DataArray(data,
                         dims=datasets[0].dims,
                         # coords=datasets[0].coords,
                         attrs=i)


class FakeCompositor(GenericCompositor):
    """Act as a compositor that produces fake RGB data."""

    def __call__(self, projectables, nonprojectables=None, **kwargs):
        """Produce test compositor data depending on modifiers and input data provided."""
        if projectables:
            projectables = self.match_data_arrays(projectables)
        if nonprojectables:
            self.match_data_arrays(nonprojectables)
        info = self.attrs.copy()
        if self.attrs["name"] in ("comp14", "comp26"):
            # used as a test when composites update the dataset id with
            # information from prereqs
            info["resolution"] = 555
        if self.attrs["name"] in ("comp24", "comp25"):
            # other composites that copy the resolution from inputs
            info["resolution"] = projectables[0].attrs.get("resolution")
        if len(projectables) != len(self.attrs["prerequisites"]):
            raise ValueError("Not enough prerequisite datasets passed")

        info.update(kwargs)
        if projectables:
            info["area"] = projectables[0].attrs["area"]
            dim_sizes = projectables[0].sizes
        else:
            # static_image
            dim_sizes = {"y": 4, "x": 5}
        return DataArray(data=da.zeros((dim_sizes["y"], dim_sizes["x"], 3)),
                         attrs=info,
                         dims=["y", "x", "bands"],
                         coords={"bands": ["R", "G", "B"]})


class FakeFileHandler(BaseFileHandler):
    """Fake file handler to be used by test readers."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialize file handler and accept all keyword arguments."""
        self.kwargs = kwargs
        super().__init__(filename, filename_info, filetype_info)

    @property
    def start_time(self):
        """Get static start time datetime object."""
        return FAKE_FILEHANDLER_START

    @property
    def end_time(self):
        """Get static end time datetime object."""
        return FAKE_FILEHANDLER_END

    @property
    def sensor_names(self):
        """Get sensor name from filetype configuration."""
        sensor = self.filetype_info.get("sensor", "fake_sensor")
        return {sensor}

    def get_dataset(self, data_id: DataID, ds_info: dict):
        """Get fake DataArray for testing."""
        if data_id["name"] == "ds9_fail_load":
            raise KeyError("Can't load '{}' because it is supposed to "
                           "fail.".format(data_id["name"]))
        attrs = data_id.to_dict()
        attrs.update(ds_info)
        attrs["sensor"] = self.filetype_info.get("sensor", "fake_sensor")
        attrs["platform_name"] = "fake_platform"
        attrs["start_time"] = self.start_time
        attrs["end_time"] = self.end_time
        res = attrs.get("resolution", 250)
        rows = cols = {
            250: 20,
            500: 10,
            1000: 5,
        }.get(res, 5)
        return DataArray(data=da.zeros((rows, cols)),
                         attrs=attrs,
                         dims=["y", "x"])

    def available_datasets(self, configured_datasets=None):
        """Report YAML datasets available unless 'not_available' is specified during creation."""
        not_available_names = self.kwargs.get("not_available", [])
        for is_avail, ds_info in (configured_datasets or []):
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info
                continue
            ft_matches = self.file_type_matches(ds_info["file_type"])
            if not ft_matches:
                yield None, ds_info
                continue
            # mimic what happens when a reader "knows" about one variable
            # but the files loaded don't have that variable
            is_avail = ds_info["name"] not in not_available_names
            yield is_avail, ds_info


class CustomScheduler:
    """Scheduler raising an exception if data are computed too many times."""

    def __init__(self, max_computes: int = 1):
        """Set starting and maximum compute counts."""
        self.max_computes = max_computes
        self.total_computes = 0

    def __call__(self, dsk, keys, **kwargs):
        """Compute dask task and keep track of number of times we do so."""
        import dask
        self.total_computes += 1
        if self.total_computes > self.max_computes:
            raise RuntimeError("Too many dask computations were scheduled: "
                               "{}".format(self.total_computes))

        return dask.get(dsk, keys, **kwargs)


@contextmanager
def assert_maximum_dask_computes(max_computes=1):
    """Context manager to make sure dask computations are not executed more than ``max_computes`` times."""
    import dask
    with dask.config.set(scheduler=CustomScheduler(max_computes=max_computes)) as new_config:
        yield new_config


def make_fake_scene(
        content_dict: Mapping,
        daskify: bool = False,
        area: bool | BaseDefinition = True,
        common_attrs: dict | None = None,
) -> Scene:
    """Create a fake Scene.

    Create a fake Scene object from fake data.  Data are provided in
    the ``content_dict`` argument.  In ``content_dict``, keys should be
    strings or DataID, and values may be either numpy.ndarray
    or xarray.DataArray, in either case with exactly two dimensions.
    The function will convert each of the numpy.ndarray objects into
    an xarray.DataArray and assign those as datasets to a Scene object.
    A fake AreaDefinition will be assigned for each array, unless disabled
    by passing ``area=False``.  When areas are automatically generated,
    arrays with the same shape will get the same area.

    This function is exclusively intended for testing purposes.

    If regular ndarrays are passed and the keyword argument daskify is
    True, DataArrays will be created as dask arrays.  If False (default),
    regular DataArrays will be created.  When the user passes xarray.DataArray
    objects then this flag has no effect.

    Args:
        content_dict: Mapping where keys correspond to objects
            accepted by ``Scene.__setitem__``, i.e. strings or DataID,
            and values may be either ``numpy.ndarray`` or
            ``xarray.DataArray``.
        daskify: optional, to use dask when converting
            ``numpy.ndarray`` to ``xarray.DataArray``.  No effect when the
            values in ``content_dict`` are already ``xarray.DataArray``.
        area: Can be ``True``, ``False``, or an
            instance of ``pyresample.geometry.BaseDefinition`` such as
            ``AreaDefinition`` or ``SwathDefinition``.  If ``True``, which is
            the default, automatically generate areas with the name "test-area".
            If ``False``, values will not have assigned areas.  If an instance
            of ``pyresample.geometry.BaseDefinition``, those instances will be
            used for all generated fake datasets.  Warning: Passing an area as
            a string (``area="germ"``) is not supported.
        common_attrs: optional, additional attributes that will
            be added to every dataset in the scene.

    Returns:
        Scene object with datasets corresponding to content_dict.
    """
    if common_attrs is None:
        common_attrs = {}
    sc = Scene()
    for (did, arr) in content_dict.items():
        extra_attrs = common_attrs.copy()
        if area:
            extra_attrs["area"] = _get_fake_scene_area(arr, area)
        sc[did] = _get_did_for_fake_scene(area, arr, extra_attrs, daskify)
    return sc


def _get_fake_scene_area(arr, area):
    """Get area for fake scene.  Helper for make_fake_scene."""
    if isinstance(area, BaseDefinition):
        return area
    return create_area_def(
        "test-area",
        {"proj": "eqc", "lat_ts": 0, "lat_0": 0, "lon_0": 0,
         "x_0": 0, "y_0": 0, "ellps": "sphere", "units": "m",
         "no_defs": None, "type": "crs"},
        units="m",
        shape=arr.shape,
        resolution=1000,
        center=(0, 0))


def _get_did_for_fake_scene(area, arr, extra_attrs, daskify):
    """Add instance to fake scene.  Helper for make_fake_scene."""
    from satpy.coords import add_crs_xy_coords
    if isinstance(arr, DataArray):
        new = arr.copy()  # don't change attributes of input
        new.attrs.update(extra_attrs)
    else:
        if daskify:
            arr = da.from_array(arr)
        new = DataArray(
                arr,
                dims=("y", "x"),
                attrs=extra_attrs)
    if area:
        new = add_crs_xy_coords(new, extra_attrs["area"])
    return new


def assert_attrs_equal(attrs, attrs_exp, tolerance=0):
    """Test that attributes are equal.

    Walks dictionary recursively. Numerical attributes are compared with
    the given relative tolerance.
    """
    keys_diff = set(attrs).difference(set(attrs_exp))
    assert not keys_diff, "Different set of keys: {}".format(keys_diff)
    for key in attrs_exp:
        err_msg = "Attribute {} does not match expectation".format(key)
        if isinstance(attrs[key], dict):
            assert_attrs_equal(attrs[key], attrs_exp[key], tolerance)
        else:
            try:
                np.testing.assert_allclose(
                    attrs[key],
                    attrs_exp[key],
                    rtol=tolerance,
                    err_msg=err_msg
                )
            except TypeError:
                assert attrs[key] == attrs_exp[key], err_msg


def assert_dict_array_equality(d1, d2):
    """Check that dicts containing arrays are equal."""
    assert set(d1.keys()) == set(d2.keys())
    for key, val1 in d1.items():
        val2 = d2[key]
        compare_func = _compare_numpy_array if isinstance(val1, np.ndarray) else _compare_nonarray
        compare_func(val1, val2)


def _compare_numpy_array(val1: np.ndarray, val2: np.ndarray) -> None:
    np.testing.assert_array_equal(val1, val2)
    assert val1.dtype == val2.dtype


def _compare_nonarray(val1: Any, val2: Any) -> None:
    assert val1 == val2
    if isinstance(val1, (np.floating, np.integer, np.bool_)):
        assert isinstance(val2, np.generic)
        assert val1.dtype == val2.dtype


def skip_numba_unstable_if_missing():
    """Determine if numba-based tests should be skipped during unstable CI tests.

    If numba fails to import it could be because numba is not compatible with
    a newer version of numpy. This is very likely to happen in the
    unstable/experimental CI environment. This function returns ``True`` if
    numba-based tests should be skipped if ``numba`` could not
    be imported *and* we're in the unstable environment. We determine if we're
    in this CI environment by looking for the ``UNSTABLE="1"``
    environment variable.

    """
    try:
        import numba
    except ImportError:
        numba = None

    return numba is None and os.environ.get("UNSTABLE", "0") in ("1", "true")



#!/usr/bin/env python
# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Shared objects of the various reader classes."""

from __future__ import annotations

from typing import Any

from satpy.utils import _import_and_warn_new_location

IMPORT_PATHS = {
    "FSFile": "satpy.readers.core.remote",
    "open_file_or_filename": "satpy.readers.core.remote",
    "group_files": "satpy.readers.core.grouping",
    "find_files_and_readers": "satpy.readers.core.grouping",
    "read_reader_config": "satpy.readers.core.config",
    "configs_for_reader": "satpy.readers.core.config",
    "available_readers": "satpy.readers.core.config",
    "get_valid_reader_names": "satpy.readers.core.config",
    "OLD_READER_NAMES": "satpy.readers.core.config",
    "PENDING_OLD_READER_NAMES": "satpy.readers.core.config",
    "load_readers": "satpy.readers.core.loading",
    "load_reader": "satpy.readers.core.loading",
}

def __getattr__(name: str) -> Any:
    new_module = IMPORT_PATHS.get(name)

    if new_module is None:
        raise AttributeError(f"module {__name__} has no attribute '{name}'")

    return _import_and_warn_new_location(new_module, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for autogenerating reader table from config files."""

from yaml import BaseLoader

from satpy.readers.core.config import available_readers


def rst_table_row(columns=None):
    """Create one row for a rst table.

    Args:
        columns (list[str]): Content of each column.

    Returns:
        str
    """
    row = "    * - {}\n".format(columns[0])
    columns = ["      - {}\n".format(col) for col in columns[1:]]
    row = row + "".join(columns)

    return row


def rst_table_header(name=None, header=None, header_rows=1, widths="auto", class_name="datatable"):
    """Create header for rst table.

    Args:
        name (str): Name of the table
        header (list[str]): Column names
        header_rows (int): Number of header rows
        widths (optional[list[int]]): Width of each column as a list. If not specified
            defaults to auto and will therefore determined by the backend
            (see <https://docutils.sourceforge.io/docs/ref/rst/directives.html#table>)
        class_name (str): The CSS class name for the table. A corresponding js function should be in main.js in
            in the "statis" directory.

    Returns:
        str
    """
    if isinstance(widths, list):
        widths = " ".join([str(w) for w in widths])

    header = rst_table_row(header)

    table_header = (f".. list-table:: {name}\n"
                    f"    :header-rows: {header_rows}\n"
                    f"    :widths: {widths}\n"
                    f"    :class: {class_name}\n\n"
                    f"{header}")

    return table_header


def generate_reader_table():
    """Create reader table from reader yaml config files.

    Returns:
        str
    """
    table = [rst_table_header("Satpy Readers", header=["Description", "Reader name", "Status", "fsspec support"],
                              widths="auto")]

    reader_configs = available_readers(as_dict=True, yaml_loader=BaseLoader)
    for rc in reader_configs:
        table.append(rst_table_row([rc.get("long_name", "").rstrip("\n"), rc.get("name", ""),
                                    rc.get("status", ""), rc.get("supports_fsspec", "false")]))

    return "".join(table)



#!/usr/bin/env python
# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utility functions for coordinates."""

import logging

import xarray as xr

LOG = logging.getLogger(__name__)


def add_xy_coords(data_arr, area, crs=None):
    """Assign x/y coordinates to DataArray from provided area.

    If 'x' and 'y' coordinates already exist then they will not be added.

    Args:
        data_arr (xarray.DataArray): data object to add x/y coordinates to
        area (pyresample.geometry.AreaDefinition): area providing the
            coordinate data.
        crs (pyproj.crs.CRS or None): CRS providing additional information
            about the area's coordinate reference system if available.
            Requires pyproj 2.0+.

    Returns (xarray.DataArray): Updated DataArray object

    """
    if "x" in data_arr.coords and "y" in data_arr.coords:
        # x/y coords already provided
        return data_arr
    if "x" not in data_arr.dims or "y" not in data_arr.dims:
        # no defined x and y dimensions
        return data_arr
    if not hasattr(area, "get_proj_vectors"):
        return data_arr
    x, y = area.get_proj_vectors()

    # convert to DataArrays
    y_attrs = {}
    x_attrs = {}

    _check_crs_units(crs, x_attrs, y_attrs)

    y = xr.DataArray(y, dims=("y",), attrs=y_attrs)
    x = xr.DataArray(x, dims=("x",), attrs=x_attrs)
    return data_arr.assign_coords(y=y, x=x)


def _check_crs_units(crs, x_attrs, y_attrs):
    if crs is None:
        return
    units = crs.axis_info[0].unit_name
    # fix udunits/CF standard units
    units = units.replace("metre", "meter")
    if units == "degree":
        y_attrs["units"] = "degrees_north"
        x_attrs["units"] = "degrees_east"
    else:
        y_attrs["units"] = units
        x_attrs["units"] = units


def add_crs_xy_coords(data_arr, area):
    """Add :class:`pyproj.crs.CRS` and x/y or lons/lats to coordinates.

    For SwathDefinition or GridDefinition areas this will add a
    `crs` coordinate and coordinates for the 2D arrays of `lons` and `lats`.

    For AreaDefinition areas this will add a `crs` coordinate and the
    1-dimensional `x` and `y` coordinate variables.

    Args:
        data_arr (xarray.DataArray): DataArray to add the 'crs'
            coordinate.
        area (pyresample.geometry.AreaDefinition): Area to get CRS
            information from.

    """
    from pyresample.geometry import SwathDefinition

    crs, data_arr = _add_crs(area, data_arr)

    # Add x/y coordinates if possible
    if isinstance(area, SwathDefinition):
        # add lon/lat arrays for swath definitions
        # SwathDefinitions created by Satpy should be assigning DataArray
        # objects as the lons/lats attributes so use those directly to
        # maintain original .attrs metadata (instead of converting to dask
        # array).
        lons = area.lons
        lats = area.lats
        lons.attrs.setdefault("standard_name", "longitude")
        lons.attrs.setdefault("long_name", "longitude")
        lons.attrs.setdefault("units", "degrees_east")
        lats.attrs.setdefault("standard_name", "latitude")
        lats.attrs.setdefault("long_name", "latitude")
        lats.attrs.setdefault("units", "degrees_north")
        # See https://github.com/pydata/xarray/issues/3068
        # data_arr = data_arr.assign_coords(longitude=lons, latitude=lats)
    else:
        # Gridded data (AreaDefinition/StackedAreaDefinition)
        data_arr = add_xy_coords(data_arr, area, crs=crs)
    return data_arr


def _add_crs(area, data_arr):
    from pyproj import CRS

    if hasattr(area, "crs"):
        crs = area.crs
    else:
        # default lat/lon projection
        latlon_proj = "+proj=latlong +datum=WGS84 +ellps=WGS84"
        proj_str = getattr(area, "proj_str", latlon_proj)
        crs = CRS.from_string(proj_str)
    data_arr = data_arr.assign_coords(crs=crs)
    return crs, data_arr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The vii_l2_nc reader tests package for VII/METimage water vapour products."""

import datetime
import os
import unittest
import uuid

import dask.array as da
import numpy as np
import xarray as xr
from netCDF4 import Dataset

from satpy.readers.vii_l2_nc import ViiL2NCFileHandler

TEST_FILE = "test_file_vii_wv_nc.nc"


class TestViiL2NCFileHandler(unittest.TestCase):
    """Test the ViiL2NCFileHandler reader."""

    def setUp(self):
        """Set up the test."""
        # Easiest way to test the reader is to create a test netCDF file on the fly
        # uses a UUID to avoid permission conflicts during execution of tests in parallel
        self.test_file_name = TEST_FILE + str(uuid.uuid1()) + ".nc"

        with Dataset(self.test_file_name, "w") as nc:
            # Create data group
            g1 = nc.createGroup("data")

            # Add dimensions to data group
            g1.createDimension("num_points_act", 100)
            g1.createDimension("num_points_alt", 10)

            # Create measurement_data group
            g1_2 = g1.createGroup("measurement_data")

            # Add variables to data/measurement_data group
            delta_lat = g1_2.createVariable("delta_lat", np.float32, dimensions=("num_points_alt", "num_points_act"))
            delta_lat[:] = 0.1

        self.reader = ViiL2NCFileHandler(
            filename=self.test_file_name,
            filename_info={
                "creation_time": datetime.datetime(year=2017, month=9, day=22,
                                                   hour=22, minute=40, second=10),
                "sensing_start_time": datetime.datetime(year=2017, month=9, day=20,
                                                        hour=12, minute=30, second=30),
                "sensing_end_time": datetime.datetime(year=2017, month=9, day=20,
                                                      hour=18, minute=30, second=50)
            },
            filetype_info={}
        )

    def tearDown(self):
        """Remove the previously created test file."""
        # Catch Windows PermissionError for removing the created test file.
        try:
            os.remove(self.test_file_name)
        except OSError:
            pass

    def test_functions(self):
        """Test the functions."""
        # Checks that the _perform_orthorectification function is correctly executed
        variable = xr.DataArray(
            dims=("num_points_alt", "num_points_act"),
            name="test_name",
            attrs={
                "key_1": "value_1",
                "key_2": "value_2"
            },
            data=da.from_array(np.ones((10, 100)))
        )
        orthorect_variable = self.reader._perform_orthorectification(variable, "data/measurement_data/delta_lat")

        expected_values = 1.1 * np.ones((10, 100))
        assert np.allclose(orthorect_variable.values, expected_values)
        assert orthorect_variable.attrs["key_1"] == "value_1"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Geostationary project utility module tests package."""

import unittest

import numpy as np

from satpy.readers.core._geos_area import (
    get_area_definition,
    get_area_extent,
    get_geos_area_naming,
    get_resolution_and_unit_strings,
    get_xy_from_linecol,
    sampling_to_lfac_cfac,
)


class TestGEOSProjectionUtil(unittest.TestCase):
    """Tests for the area utilities."""

    def make_pdict_ext(self, typ, scan):
        """Create a dictionary and extents to use in testing."""
        if typ == 1:  # Fulldisk
            pdict = {"a": 6378169.00,
                     "b": 6356583.80,
                     "h": 35785831.00,
                     "ssp_lon": 0.0,
                     "nlines": 3712,
                     "ncols": 3712,
                     "a_name": "geostest",
                     "a_desc": "test area",
                     "p_id": "test_area",
                     "cfac": -13642337,
                     "lfac": -13642337,
                     "coff": 1856}
            if scan == "N2S":
                pdict["scandir"] = "N2S"
                pdict["loff"] = 1856
                extent = (5567248.28340708,
                          5567248.28340708,
                          -5570248.686685662,
                          -5570248.686685662)

            if scan == "S2N":
                pdict["scandir"] = "S2N"
                pdict["loff"] = -1856
                extent = (5567248.28340708,
                          5570248.686685662,
                          -5570248.686685662,
                          -5567248.28340708)

        if typ == 2:  # One sector
            pdict = {"a": 6378169.00,
                     "b": 6356583.80,
                     "h": 35785831.00,
                     "ssp_lon": 0.0,
                     "nlines": 464,
                     "ncols": 3712,
                     "a_name": "geostest",
                     "a_desc": "test area",
                     "p_id": "test_area",
                     "cfac": -13642337,
                     "lfac": -13642337,
                     "coff": 1856}
            if scan == "N2S":
                pdict["scandir"] = "N2S"
                pdict["loff"] = 464
                extent = (5567248.28340708,
                          1390686.9196223018,
                          -5570248.686685662,
                          -1500.2016392905093)

            if scan == "S2N":
                pdict["scandir"] = "S2N"
                pdict["loff"] = 464
                extent = (5567248.28340708,
                          -1390686.9196223018,
                          -5570248.686685662,
                          -2782874.0408838945)

        return pdict, extent

    def test_geos_area(self):
        """Test area extent calculation with N->S scan then S->N scan."""
        # North -> South full disk
        pdict, extent = self.make_pdict_ext(1, "N2S")
        aex = get_area_extent(pdict)
        np.testing.assert_allclose(aex, extent)

        # South -> North full disk
        pdict, extent = self.make_pdict_ext(1, "S2N")
        aex = get_area_extent(pdict)
        np.testing.assert_allclose(aex, extent)

        # North -> South one sector
        pdict, extent = self.make_pdict_ext(2, "N2S")
        aex = get_area_extent(pdict)
        np.testing.assert_allclose(aex, extent)

        # South -> North one sector
        pdict, extent = self.make_pdict_ext(2, "S2N")
        aex = get_area_extent(pdict)
        np.testing.assert_allclose(aex, extent)

    def test_get_xy_from_linecol(self):
        """Test the scan angle calculation."""
        pdict, extent = self.make_pdict_ext(1, "S2N")
        good_xy = [0.2690166648133674, -10.837528496767087]
        factors = (pdict["lfac"], pdict["cfac"])
        offsets = (pdict["loff"], pdict["coff"])
        x, y = get_xy_from_linecol(400, 1800, offsets, factors)
        np.testing.assert_approx_equal(x, good_xy[0])
        np.testing.assert_approx_equal(y, good_xy[1])

        pdict, extent = self.make_pdict_ext(2, "N2S")
        good_xy = [0.2690166648133674, 0.30744761692956274]
        factors = (pdict["lfac"], pdict["cfac"])
        offsets = (pdict["loff"], pdict["coff"])
        x, y = get_xy_from_linecol(400, 1800, offsets, factors)
        np.testing.assert_approx_equal(x, good_xy[0])
        np.testing.assert_approx_equal(y, good_xy[1])

    def test_get_area_definition(self):
        """Test the retrieval of the area definition."""
        from pyproj import CRS

        pdict, extent = self.make_pdict_ext(1, "N2S")
        good_res = (-3000.4032785810186, -3000.4032785810186)

        a_def = get_area_definition(pdict, extent)
        assert a_def.area_id == pdict["a_name"]
        assert a_def.resolution == good_res

        expected_crs = CRS(dict(proj="geos", units="m", a=6378169, b=6356583.8, h=35785831))
        assert a_def.crs == expected_crs

    def test_sampling_to_lfac_cfac(self):
        """Test conversion from angular sampling to line/column offset."""
        lfac = 13642337  # SEVIRI LFAC
        sampling = np.deg2rad(2 ** 16 / lfac)
        np.testing.assert_allclose(sampling_to_lfac_cfac(sampling), lfac)

    def test_get_geos_area_naming(self):
        """Test the geos area naming function."""
        input_dict = {"platform_name": "testplatform",
                      "instrument_name": "testinstrument",
                      "resolution": 1000,
                      "service_name": "testservicename",
                      "service_desc": "testdesc"}

        output_dict = get_geos_area_naming(input_dict)

        assert output_dict["area_id"] == "testplatform_testinstrument_testservicename_1km"
        assert output_dict["description"] == "TESTPLATFORM TESTINSTRUMENT testdesc area definition with 1 km resolution"

    def test_get_resolution_and_unit_strings_in_km(self):
        """Test the resolution and unit strings function for a km resolution."""
        out = get_resolution_and_unit_strings(1000)
        assert out["value"] == "1"
        assert out["unit"] == "km"

    def test_get_resolution_and_unit_strings_in_m(self):
        """Test the resolution and unit strings function for a m resolution."""
        out = get_resolution_and_unit_strings(500)
        assert out["value"] == "500"
        assert out["unit"] == "m"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2012-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for aapp level 1b data.

Options for loading:

 - pre_launch_coeffs (False): use pre-launch coefficients if True, operational
   otherwise (if available).

https://nwp-saf.eumetsat.int/site/download/documentation/aapp/NWPSAF-MF-UD-003_Formats_v8.0.pdf
"""

import datetime as dt
import functools
import logging

import dask.array as da
import numpy as np
import xarray as xr
from dask import delayed

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_chunk_size_limit

CHANNEL_DTYPE = np.float32


def get_avhrr_lac_chunks(shape, dtype):
    """Get chunks from a given shape adapted for full-resolution AVHRR data."""
    limit = get_chunk_size_limit(dtype)
    return da.core.normalize_chunks(("auto", 2048), shape=shape, limit=limit, dtype=dtype)


def get_aapp_chunks(shape):
    """Get chunks from a given shape adapted for AAPP data."""
    return get_avhrr_lac_chunks(shape, dtype=CHANNEL_DTYPE)


logger = logging.getLogger(__name__)

AVHRR_CHANNEL_NAMES = ["1", "2", "3a", "3b", "4", "5"]

AVHRR_ANGLE_NAMES = ["sensor_zenith_angle",
                     "solar_zenith_angle",
                     "sun_sensor_azimuth_difference_angle"]

AVHRR_PLATFORM_IDS2NAMES = {4: "NOAA-15",
                            2: "NOAA-16",
                            6: "NOAA-17",
                            7: "NOAA-18",
                            8: "NOAA-19",
                            11: "Metop-B",
                            12: "Metop-A",
                            13: "Metop-C",
                            14: "Metop simulator"}


def create_xarray(arr):
    """Create an `xarray.DataArray`."""
    res = xr.DataArray(arr, dims=["y", "x"])
    return res


class AAPPL1BaseFileHandler(BaseFileHandler):
    """A base file handler for the AAPP level-1 formats."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize AAPP level-1 file handler object."""
        super().__init__(filename, filename_info, filetype_info)

        self.channels = None
        self.units = None
        self.sensor = "unknown"

        self._data = None
        self._header = None
        self.area = None

        self._channel_names = []
        self._angle_names = []

    def _set_filedata_layout(self):
        """Set the file data type/layout."""
        self._header_offset = 0
        self._scan_type = np.dtype([("siteid", "<i2")])
        self._header_type = np.dtype([("siteid", "<i2")])

    @property
    def start_time(self):
        """Get the time of the first observation."""
        return dt.datetime(self._data["scnlinyr"][0], 1, 1) + dt.timedelta(
            days=int(self._data["scnlindy"][0]) - 1,
            milliseconds=int(self._data["scnlintime"][0]))

    @property
    def end_time(self):
        """Get the time of the final observation."""
        return dt.datetime(self._data["scnlinyr"][-1], 1, 1) + dt.timedelta(
            days=int(self._data["scnlindy"][-1]) - 1,
            milliseconds=int(self._data["scnlintime"][-1]))

    def _update_dataset_attributes(self, dataset, key, info):
        dataset.attrs.update({"platform_name": self.platform_name,
                              "sensor": self.sensor})
        dataset.attrs.update(key.to_dict())
        for meta_key in ("standard_name", "units"):
            if meta_key in info:
                dataset.attrs.setdefault(meta_key, info[meta_key])

    def _get_platform_name(self, platform_names_lookup):
        """Get the platform name from the file header."""
        self.platform_name = platform_names_lookup.get(self._header["satid"][0], None)
        if self.platform_name is None:
            raise ValueError("Unsupported platform ID: %d" % self.header["satid"])

    def read(self):
        """Read the data."""
        tic = dt.datetime.now()
        header = np.memmap(self.filename, dtype=self._header_type, mode="r", shape=(1, ))
        data = np.memmap(self.filename, dtype=self._scan_type, offset=self._header_offset, mode="r")
        logger.debug("Reading time %s", str(dt.datetime.now() - tic))

        self._header = header
        self._data = data

    def _calibrate_active_channel_data(self, key):
        """Calibrate active channel data only."""
        raise NotImplementedError("This should be implemented in the sub class")

    def get_dataset(self, key, info):
        """Get a dataset from the file."""
        if key["name"] in self._channel_names:
            dataset = self._calibrate_active_channel_data(key)
            if dataset is None:
                return None
        elif key["name"] in ["longitude", "latitude"]:
            dataset = self.navigate(key["name"])
            dataset.attrs = info
        elif key["name"] in self._angle_names:
            dataset = self.get_angles(key["name"])
        else:
            raise ValueError("Not a supported dataset: %s", key["name"])

        self._update_dataset_attributes(dataset, key, info)
        return dataset


class AVHRRAAPPL1BFile(AAPPL1BaseFileHandler):
    """Reader for AVHRR L1B files created from the AAPP software."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize object information by reading the input file."""
        super(AVHRRAAPPL1BFile, self).__init__(filename, filename_info,
                                               filetype_info)

        self.channels = {i: None for i in AVHRR_CHANNEL_NAMES}
        self.units = {i: "counts" for i in AVHRR_CHANNEL_NAMES}

        self._is3b = None
        self._is3a = None
        self._channel_names = AVHRR_CHANNEL_NAMES
        self._angle_names = AVHRR_ANGLE_NAMES

        self._set_filedata_layout()
        self.read()

        self.active_channels = self._get_active_channels()

        self._get_platform_name(AVHRR_PLATFORM_IDS2NAMES)
        self.sensor = "avhrr-3"

        self._get_all_interpolated_angles = functools.lru_cache(maxsize=10)(
            self._get_all_interpolated_angles_uncached
        )
        self._get_all_interpolated_coordinates = functools.lru_cache(maxsize=10)(
            self._get_all_interpolated_coordinates_uncached
        )

    def _set_filedata_layout(self):
        """Set the file data type/layout."""
        self._header_offset = 22016
        self._scan_type = _SCANTYPE
        self._header_type = _HEADERTYPE

    def _get_active_channels(self):
        status = self._get_channel_binary_status_from_header()
        return self._convert_binary_channel_status_to_activation_dict(status)

    def _calibrate_active_channel_data(self, key):
        """Calibrate active channel data only."""
        if self.active_channels[key["name"]]:
            return self.calibrate(key)
        return None

    def _get_channel_binary_status_from_header(self):
        status = self._header["inststat1"].item()
        change_line = self._header["statchrecnb"]
        if change_line > 0:
            status |= self._header["inststat2"].item()
        return status

    @staticmethod
    def _convert_binary_channel_status_to_activation_dict(status):
        bits_channels = ((13, "1"),
                         (12, "2"),
                         (11, "3a"),
                         (10, "3b"),
                         (9, "4"),
                         (8, "5"))
        activated = dict()
        for bit, channel_name in bits_channels:
            activated[channel_name] = bool(status >> bit & 1)
        return activated

    def available_datasets(self, configured_datasets=None):
        """Get the available datasets."""
        for _, mda in configured_datasets:
            if mda["name"] in self._channel_names:
                yield self.active_channels[mda["name"]], mda
            else:
                yield True, mda

    def get_angles(self, angle_id):
        """Get sun-satellite viewing angles."""
        sunz, satz, azidiff = self._get_all_interpolated_angles()
        name_to_variable = dict(zip(self._angle_names, (satz, sunz, azidiff)))
        return create_xarray(name_to_variable[angle_id])

    def _get_all_interpolated_angles_uncached(self):
        sunz40km, satz40km, azidiff40km = self._get_tiepoint_angles_in_degrees()
        return self._interpolate_arrays(sunz40km, satz40km, azidiff40km)

    def _get_tiepoint_angles_in_degrees(self):
        angles = self._data["ang"].astype(np.float32)
        sunz40km = angles[:, :, 0] * 1e-2
        satz40km = angles[:, :, 1] * 1e-2
        azidiff40km = angles[:, :, 2] * 1e-2
        return sunz40km, satz40km, azidiff40km

    def _interpolate_arrays(self, *input_arrays, geolocation=False):
        lines = input_arrays[0].shape[0]
        try:
            interpolator = self._create_40km_interpolator(lines, *input_arrays, geolocation=geolocation)
        except ImportError:
            logger.warning("Could not interpolate, python-geotiepoints missing.")
            output_arrays = input_arrays
        else:
            output_delayed = delayed(interpolator.interpolate, nout=3)()
            output_arrays = [da.from_delayed(out_array, (lines, 2048), in_array.dtype)
                             for in_array, out_array in zip(input_arrays, output_delayed)]
        return output_arrays

    @staticmethod
    def _create_40km_interpolator(lines, *arrays_40km, geolocation=False):
        if geolocation:
            # Slower but accurate at datum line
            from geotiepoints.geointerpolator import GeoInterpolator as Interpolator
        else:
            from geotiepoints.interpolator import Interpolator
        cols40km = np.arange(24, 2048, 40)
        cols1km = np.arange(2048)
        rows40km = np.arange(lines)
        rows1km = np.arange(lines)
        along_track_order = 1
        cross_track_order = 3
        satint = Interpolator(
            arrays_40km, (rows40km, cols40km),
            (rows1km, cols1km), along_track_order, cross_track_order)
        return satint

    def navigate(self, coordinate_id):
        """Get the longitudes and latitudes of the scene."""
        lons, lats = self._get_all_interpolated_coordinates()
        if coordinate_id == "longitude":
            return create_xarray(lons)
        if coordinate_id == "latitude":
            return create_xarray(lats)

        raise KeyError("Coordinate {} unknown.".format(coordinate_id))

    def _get_all_interpolated_coordinates_uncached(self):
        lons40km, lats40km = self._get_coordinates_in_degrees()
        return self._interpolate_arrays(lons40km, lats40km, geolocation=True)

    def _get_coordinates_in_degrees(self):
        position_data = self._data["pos"].astype(np.float32)
        lons40km = position_data[:, :, 1] * 1e-4
        lats40km = position_data[:, :, 0] * 1e-4

        return lons40km, lats40km

    def calibrate(self,
                  dataset_id,
                  pre_launch_coeffs=False,
                  calib_coeffs=None):
        """Calibrate the data."""
        if calib_coeffs is None:
            calib_coeffs = {}

        units = {"reflectance": "%",
                 "brightness_temperature": "K",
                 "counts": "",
                 "radiance": "W*m-2*sr-1*cm ?"}

        if dataset_id["name"] in ("3a", "3b") and self._is3b is None:
            # Is it 3a or 3b:
            line_chunks = get_aapp_chunks((self._data.shape[0], 2048))[0]
            self._is3a = da.bitwise_and(da.from_array(self._data["scnlinbit"],
                                                      chunks=line_chunks), 3) == 0
            self._is3b = da.bitwise_and(da.from_array(self._data["scnlinbit"],
                                                      chunks=line_chunks), 3) == 1

        try:
            vis_idx = ["1", "2", "3a"].index(dataset_id["name"])
            ir_idx = None
        except ValueError:
            vis_idx = None
            ir_idx = ["3b", "4", "5"].index(dataset_id["name"])

        mask = True
        if vis_idx is not None:
            coeffs = calib_coeffs.get("ch" + dataset_id["name"])
            if dataset_id["name"] == "3a":
                mask = self._is3a[:, None]
            ds = create_xarray(
                _vis_calibrate(self._data,
                               vis_idx,
                               dataset_id["calibration"],
                               pre_launch_coeffs,
                               coeffs,
                               mask=mask))
        else:
            if dataset_id["name"] == "3b":
                mask = self._is3b[:, None]
            ds = create_xarray(
                _ir_calibrate(self._header,
                              self._data,
                              ir_idx,
                              dataset_id["calibration"],
                              mask=mask))

        ds.attrs["units"] = units[dataset_id["calibration"]]
        ds.attrs.update(dataset_id._asdict())
        return ds


# AAPP 1b header

_HEADERTYPE = np.dtype([("siteid", "S3"),
                        ("blank", "S1"),
                        ("l1bversnb", "<i2"),
                        ("l1bversyr", "<i2"),
                        ("l1bversdy", "<i2"),
                        ("reclg", "<i2"),
                        ("blksz", "<i2"),
                        ("hdrcnt", "<i2"),
                        ("filler0", "S6"),
                        ("dataname", "S42"),
                        ("prblkid", "S8"),
                        ("satid", "<i2"),
                        ("instid", "<i2"),
                        ("datatype", "<i2"),
                        ("tipsrc", "<i2"),
                        ("startdatajd", "<i4"),
                        ("startdatayr", "<i2"),
                        ("startdatady", "<i2"),
                        ("startdatatime", "<i4"),
                        ("enddatajd", "<i4"),
                        ("enddatayr", "<i2"),
                        ("enddatady", "<i2"),
                        ("enddatatime", "<i4"),
                        ("cpidsyr", "<i2"),
                        ("cpidsdy", "<i2"),
                        ("filler1", "S8"),
                        # data set quality indicators
                        ("inststat1", "<i4"),
                        ("filler2", "S2"),
                        ("statchrecnb", "<i2"),
                        ("inststat2", "<i4"),
                        ("scnlin", "<i2"),
                        ("callocscnlin", "<i2"),
                        ("misscnlin", "<i2"),
                        ("datagaps", "<i2"),
                        ("okdatafr", "<i2"),
                        ("pacsparityerr", "<i2"),
                        ("auxsyncerrsum", "<i2"),
                        ("timeseqerr", "<i2"),
                        ("timeseqerrcode", "<i2"),
                        ("socclockupind", "<i2"),
                        ("locerrind", "<i2"),
                        ("locerrcode", "<i2"),
                        ("pacsstatfield", "<i2"),
                        ("pacsdatasrc", "<i2"),
                        ("filler3", "S4"),
                        ("spare1", "S8"),
                        ("spare2", "S8"),
                        ("filler4", "S10"),
                        # Calibration
                        ("racalind", "<i2"),
                        ("solarcalyr", "<i2"),
                        ("solarcaldy", "<i2"),
                        ("pcalalgind", "<i2"),
                        ("pcalalgopt", "<i2"),
                        ("scalalgind", "<i2"),
                        ("scalalgopt", "<i2"),
                        ("irttcoef", "<i2", (4, 6)),
                        ("filler5", "<i4", (2, )),
                        # radiance to temperature conversion
                        ("albcnv", "<i4", (2, 3)),
                        ("radtempcnv", "<i4", (3, 3)),
                        ("filler6", "<i4", (3, )),
                        # Navigation
                        ("modelid", "S8"),
                        ("nadloctol", "<i2"),
                        ("locbit", "<i2"),
                        ("filler7", "S2"),
                        ("rollerr", "<i2"),
                        ("pitcherr", "<i2"),
                        ("yawerr", "<i2"),
                        ("epoyr", "<i2"),
                        ("epody", "<i2"),
                        ("epotime", "<i4"),
                        ("smaxis", "<i4"),
                        ("eccen", "<i4"),
                        ("incli", "<i4"),
                        ("argper", "<i4"),
                        ("rascnod", "<i4"),
                        ("manom", "<i4"),
                        ("xpos", "<i4"),
                        ("ypos", "<i4"),
                        ("zpos", "<i4"),
                        ("xvel", "<i4"),
                        ("yvel", "<i4"),
                        ("zvel", "<i4"),
                        ("earthsun", "<i4"),
                        ("filler8", "S16"),
                        # analog telemetry conversion
                        ("pchtemp", "<i2", (5, )),
                        ("reserved1", "<i2"),
                        ("pchtempext", "<i2", (5, )),
                        ("reserved2", "<i2"),
                        ("pchpow", "<i2", (5, )),
                        ("reserved3", "<i2"),
                        ("rdtemp", "<i2", (5, )),
                        ("reserved4", "<i2"),
                        ("bbtemp1", "<i2", (5, )),
                        ("reserved5", "<i2"),
                        ("bbtemp2", "<i2", (5, )),
                        ("reserved6", "<i2"),
                        ("bbtemp3", "<i2", (5, )),
                        ("reserved7", "<i2"),
                        ("bbtemp4", "<i2", (5, )),
                        ("reserved8", "<i2"),
                        ("eleccur", "<i2", (5, )),
                        ("reserved9", "<i2"),
                        ("motorcur", "<i2", (5, )),
                        ("reserved10", "<i2"),
                        ("earthpos", "<i2", (5, )),
                        ("reserved11", "<i2"),
                        ("electemp", "<i2", (5, )),
                        ("reserved12", "<i2"),
                        ("chtemp", "<i2", (5, )),
                        ("reserved13", "<i2"),
                        ("bptemp", "<i2", (5, )),
                        ("reserved14", "<i2"),
                        ("mhtemp", "<i2", (5, )),
                        ("reserved15", "<i2"),
                        ("adcontemp", "<i2", (5, )),
                        ("reserved16", "<i2"),
                        ("d4bvolt", "<i2", (5, )),
                        ("reserved17", "<i2"),
                        ("d5bvolt", "<i2", (5, )),
                        ("reserved18", "<i2"),
                        ("bbtempchn3b", "<i2", (5, )),
                        ("reserved19", "<i2"),
                        ("bbtempchn4", "<i2", (5, )),
                        ("reserved20", "<i2"),
                        ("bbtempchn5", "<i2", (5, )),
                        ("reserved21", "<i2"),
                        ("refvolt", "<i2", (5, )),
                        ("reserved22", "<i2"), ])

# AAPP 1b scanline

_SCANTYPE = np.dtype([("scnlin", "<i2"),
                      ("scnlinyr", "<i2"),
                      ("scnlindy", "<i2"),
                      ("clockdrift", "<i2"),
                      ("scnlintime", "<i4"),
                      ("scnlinbit", "<i2"),
                      ("filler0", "S10"),
                      ("qualind", "<i4"),
                      ("scnlinqual", "<i4"),
                      ("calqual", "<i2", (3, )),
                      ("cbiterr", "<i2"),
                      ("filler1", "S8"),
                      # Calibration
                      ("calvis", "<i4", (3, 3, 5)),
                      ("calir", "<i4", (3, 2, 3)),
                      ("filler2", "<i4", (3, )),
                      # Navigation
                      ("navstat", "<i4"),
                      ("attangtime", "<i4"),
                      ("rollang", "<i2"),
                      ("pitchang", "<i2"),
                      ("yawang", "<i2"),
                      ("scalti", "<i2"),
                      ("ang", "<i2", (51, 3)),
                      ("filler3", "<i2", (3, )),
                      ("pos", "<i4", (51, 2)),
                      ("filler4", "<i4", (2, )),
                      ("telem", "<i2", (103, )),
                      ("filler5", "<i2"),
                      ("hrpt", "<i2", (2048, 5)),
                      ("filler6", "<i4", (2, )),
                      # tip minor frame header
                      ("tipmfhd", "<i2", (7, 5)),
                      # cpu telemetry
                      ("cputel", "S6", (2, 5)),
                      ("filler7", "<i2", (67, )), ])


def _vis_calibrate(data,
                   chn,
                   calib_type,
                   pre_launch_coeffs=False,
                   calib_coeffs=None,
                   mask=True):
    """Calibrate visible channel data.

    *calib_type* in count, reflectance, radiance.

    """
    # Calibration count to albedo, the calibration is performed separately for
    # two value ranges.
    if calib_type not in ["counts", "radiance", "reflectance"]:
        raise ValueError("Calibration " + calib_type + " unknown!")

    channel_data = data["hrpt"][:, :, chn]
    chunks = get_aapp_chunks(channel_data.shape)
    line_chunks = chunks[0]
    channel = da.from_array(channel_data, chunks=chunks)
    mask &= channel != 0

    if calib_type == "counts":
        return channel

    channel = channel.astype(CHANNEL_DTYPE)

    if calib_type == "radiance":
        logger.info("Radiances are not yet supported for " +
                    "the VIS/NIR channels!")

    if pre_launch_coeffs:
        coeff_idx = 2
    else:
        # check that coeffs are valid
        if np.all(data["calvis"][:, chn, 0, 4] == 0):
            logger.info(
                "No valid operational coefficients, fall back to pre-launch")
            coeff_idx = 2
        else:
            coeff_idx = 0

    intersection = da.from_array(data["calvis"][:, chn, coeff_idx, 4],
                                 chunks=line_chunks)

    if calib_coeffs is not None:
        logger.info("Updating from external calibration coefficients.")
        slope1 = da.from_array(calib_coeffs[0], chunks=line_chunks)
        intercept1 = da.from_array(calib_coeffs[1], chunks=line_chunks)
        slope2 = da.from_array(calib_coeffs[2], chunks=line_chunks)
        intercept2 = da.from_array(calib_coeffs[3], chunks=line_chunks)
    else:
        calvis = data["calvis"]
        slope1 = da.from_array(calvis[:, chn, coeff_idx, 0], chunks=line_chunks).astype(np.float32) * 1e-10
        intercept1 = da.from_array(calvis[:, chn, coeff_idx, 1], chunks=line_chunks).astype(np.float32) * 1e-7
        slope2 = da.from_array(calvis[:, chn, coeff_idx, 2], chunks=line_chunks).astype(np.float32) * 1e-10
        intercept2 = da.from_array(calvis[:, chn, coeff_idx, 3], chunks=line_chunks).astype(np.float32) * 1e-7

        # In the level 1b file, the visible coefficients are stored as 4-byte integers. Scaling factors then convert
        # them to real numbers which are applied to the measured counts. The coefficient is different depending on
        # whether the counts are less than or greater than the high-gain/low-gain transition value (nominally 500).
        # The slope for visible channels should always be positive (reflectance increases with count). With the
        # pre-launch coefficients the channel 2, 3a slope is always positive but with the operational coefs the stored
        # number in the high-reflectance regime overflows the maximum 2147483647, i.e. it is negative when
        # interpreted as a signed integer. So you have to modify it. Also chanel 1 is treated the same way in AAPP.
        slope2 = da.where(slope2 < 0, slope2 + 0.4294967296, slope2)

    channel = da.where(channel <= intersection[:, None],
                       channel * slope1[:, None] + intercept1[:, None],
                       channel * slope2[:, None] + intercept2[:, None])

    channel = channel.clip(min=0)

    return da.where(mask, channel, np.nan)


def _ir_calibrate(header, data, irchn, calib_type, mask=True):
    """Calibrate for IR bands.

    *calib_type* in brightness_temperature, radiance, count

    """
    channel_data = data["hrpt"][:, :, irchn + 2]
    chunks = get_aapp_chunks(channel_data.shape)
    line_chunks = chunks[0]

    count = da.from_array(channel_data, chunks=chunks)

    if calib_type == 0:
        return count

    # Mask unnaturally low values
    mask &= count != 0
    count = count.astype(CHANNEL_DTYPE)

    calir = data["calir"]
    k1_ = da.from_array(calir[:, irchn, 0, 0], chunks=line_chunks).astype(np.float32) * 1.0e-9
    k2_ = da.from_array(calir[:, irchn, 0, 1], chunks=line_chunks).astype(np.float32) * 1.0e-6
    k3_ = da.from_array(calir[:, irchn, 0, 2], chunks=line_chunks).astype(np.float32) * 1.0e-6

    # Count to radiance conversion:
    rad = k1_[:, None] * count * count + k2_[:, None] * count + k3_[:, None]

    # Suspicious lines
    mask &= ((k1_ != 0) | (k2_ != 0) | (k3_ != 0))[:, None]

    if calib_type == 2:
        mask &= rad > 0.0
        return da.where(mask, rad, np.nan)

    radtempcnv = header["radtempcnv"].astype(np.float32)

    # Central wavenumber:
    cwnum = radtempcnv[0, irchn, 0]
    if irchn == 0:
        cwnum = cwnum / 1.0e2
    else:
        cwnum = cwnum / 1.0e3
    bandcor_2 = radtempcnv[0, irchn, 1] / 1e5
    bandcor_3 = radtempcnv[0, irchn, 2] / 1e6
    return da.map_blocks(_ir_calibrate_numpy,
                         rad, mask, cwnum, bandcor_2, bandcor_3,
                         dtype=rad.dtype,
                         meta=np.ndarray((), dtype=rad.dtype))



def _ir_calibrate_numpy(rad, mask, cwnum, bandcor_2, bandcor_3):
    """Calibrate for IR bands as numpy arrays to avoid dividing by 0."""
    ir_const_1 = 1.1910659e-5
    ir_const_2 = 1.438833

    good_rad = rad[mask]
    t_planck = (ir_const_2 * cwnum) / \
               np.log(1 + ir_const_1 * cwnum * cwnum * cwnum / good_rad)

    # Band corrections applied to t_planck to get correct
    # brightness temperature for channel:
    if bandcor_2 < 0:  # Post AAPP-v4
        tb_ = bandcor_2 + bandcor_3 * t_planck
    else:  # AAPP 1 to 4
        tb_ = (t_planck - bandcor_2) / bandcor_3

    tb = np.full(rad.shape, np.float32(np.nan), dtype=np.float32)
    tb[mask] = tb_
    return tb
